#include <stdio.h>
#include <stdlib.h>     /* for malloc() and free() */
#include <string.h>     /* for memcpy() */

/* All numeric fields are in Motorola (Big Endian) order */
typedef struct {
  long signature;       /* should be 'Rand' == 0x646e6152L */
  long numframes;
  short packing;        /* only 0 is defined */
  short reserved1;      /* should be 0 */
  short width;
  short height;
  short leading;
  short fps;            /* frames per second */
  char  reserved2[16];  /* undefined */
} FilmStrip;

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

typedef struct {
  uchar r,g,b,a;
} RGBA;

typedef struct {
  uchar r,g,b;
} RGB;

typedef uchar TRIPLE[3];      /* convenient equivalent for RGB */

static short sswap(short value)
{
  ushort v = value;
  v = ( (v & 0xff) << 8 ) | ((v >> 8) & 0xff);
  return (short)v;
}

static long lswap(long value)
{
  ulong v = value;
  v =   ( (v & 0xff) << 24 )
      | ( ((v >> 8) & 0xff) << 16 )
      | ( ((v >> 16) & 0xff) << 8 )
      |   ((v >> 24) & 0xff);
  return (long)v;
}

void PCX_writeheader(FILE *fp,int width,int height,int leading,int frames);
void PCX_writeline(FILE *fp,RGB *line,int width);

int main(int argc,char *argv[])
{
  FILE *fin,*fout;
  FilmStrip fs;
  RGB *pixels;
  RGBA rgba;

  if (argc<2 || argc>3) {
    printf("FILMSTR: returns information on Adobe filmstrip files, and optionally\n"
           "writes a PCX file for use in other graphic editing programs\n\n");
    printf("Usage:\tfilmstr <input> [output]\n");
    return 1;
  } /* if */

  if ((fin=fopen(argv[1],"rb"))==NULL) {
    printf("Cannot open file \"%s\".\n",argv[1]);
    return 1;
  } /* if */

  fseek(fin, - (long)(sizeof fs), SEEK_END);
  fread(&fs, sizeof fs, 1, fin);

  if (fs.signature!=0x646e6152L) {
    printf("File \"%s\" is not a valid \"filmstrip\" file\n",argv[1]);
    return 1;
  } /* if */

  /* convert, for easier manipulation */
  fs.numframes=lswap(fs.numframes);
  fs.width=sswap(fs.width);
  fs.height=sswap(fs.height);
  fs.leading=sswap(fs.leading);
  fs.fps=sswap(fs.fps);

  printf("Packing:          %s\n",fs.packing ? "RGB" : "RGB+Alpha");
  printf("Number of frames: %ld\n",fs.numframes);
  printf("Filmstrip width:  %d\n",fs.width);
  printf("Filmstrip height: %ld\n",(fs.height+fs.leading)*fs.numframes);
  printf("Frame height:     %d\n",fs.height);
  printf("Frame leading:    %d\n",fs.leading);
  printf("Playback speed:   %d fps\n",fs.fps);

  if (argc==3) {
    long line,numlines;
    int x;
    int percent,prev_percent;

    fseek(fin,0L,SEEK_SET);
    if ((pixels=malloc(fs.width*sizeof(RGB)))==NULL) {
      printf("Cannot create file \"%s\".\n",argv[2]);
      fclose(fin);
      return 1;
    } /* if */
    if ((fout=fopen(argv[2],"wb"))==NULL) {
      printf("Cannot create file \"%s\".\n",argv[2]);
      fclose(fin);
      free(pixels);
      return 1;
    } /* if */

    PCX_writeheader(fout,fs.width,fs.height,fs.leading,(int)fs.numframes);
    prev_percent=0;
    printf("done  0%%");
    numlines=(fs.height+fs.leading)*fs.numframes;
    for (line=0; line<numlines; line++) {
      percent=(int)((100L*line)/numlines);
      if (prev_percent!=percent) {
        printf("\b\b\b%2d%%",percent);
        prev_percent=percent;
      } /* if */
      for (x=0; x<fs.width; x++) {
        fread(&rgba,4,1,fin);
        memcpy(&pixels[x],&rgba,3);
      } /* if */
      PCX_writeline(fout,pixels,fs.width);
    } /* for */
    printf("\b\b\b\b    \n");
    fflush(stdout);
    fclose(fout);
    free(pixels);
  } /* if */

  fclose(fin);
  return 0;
}

// PCX code
typedef struct {
  uchar  manufacturer;  // always 0xa
  uchar  version;
  uchar  encoding;      // always 1 (RLE)
  uchar  bitsperpixel;
  ushort x1, y1;
  ushort x2, y2;
  ushort res_x, res_y;  // horizontal and vertical display resolution
  RGB    palette[16];
  uchar  mode;          // video mode
  uchar  planes;        // number of colour planes
  ushort bytesperline;  // always an even number
  ushort photometric;   // photometric interpretation: colour or gray scale
  ushort scanlines;
  uchar  filler[56];    // filler up to 128 bytes
} PCXHDR;

void PCX_writeheader(FILE *fp,int width,int height,int leading,int frames)
{
  PCXHDR hdr;
  int bytesperline = (width + 1) & 0xfffe;

  /*
  ** Fill in the PCX header
  */
  memset(&hdr,0,128);
  hdr.manufacturer = (char)10;
  hdr.version = (char)5;
  hdr.encoding = (char)1;
  hdr.bitsperpixel = 8;
  hdr.x1 = hdr.y1 = 0;
  hdr.x2 = width - 1;
  hdr.y2 = (int)((long)(leading + height) * frames) - 1;
  hdr.res_x = 640;
  hdr.res_y = 480;
  hdr.bytesperline = bytesperline;
  hdr.mode = (char)0;
  hdr.planes = 3;
  hdr.photometric = 1;

  fwrite(&hdr, sizeof(PCXHDR), 1, fp);
}

void PCX_writeline(FILE *fp,RGB *line,int width)
{
  int plane,index,count,val;
  int bytesperline = (width + 1) & 0xfffe;
  uchar  pixel;
  TRIPLE *buffer=(TRIPLE *)line;

  for (plane=0; plane<3; plane++) {
    for (count=0; count<bytesperline; ) {
      index=count;
      pixel=buffer[index++][plane];
      while (index<bytesperline && index-count<63 && buffer[index][plane]==pixel)
        index++;
      val=(index-count);
      if (val==1) {
        if (pixel<0xc0) {
          fputc(pixel,fp);
        } else {
          fputc(0xc1,fp);   /* run with length 1 */
          fputc(pixel,fp);
        } /* if */
      } else {
        fputc(val+0xc0,fp);
        fputc(pixel,fp);
      } /* if */
      count=index;
    } /* for */
  } /* for */
}

